#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore.
"""
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Union, Set, Any, Generator, Collection
import os
import io
from warnings import warn
from collections import OrderedDict
import copy

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.list_records import ListRecords, GeneralDataFrame
from ckanapi_harvesters.builder.builder_field import BuilderField
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC
from ckanapi_harvesters.harvesters.file_formats.file_format_init import init_file_format_datastore
from ckanapi_harvesters.builder.mapper_datastore import DataSchemeConversion
from ckanapi_harvesters.builder.builder_resource import BuilderResourceABC
from ckanapi_harvesters.auxiliary.ckan_errors import DuplicateNameError
from ckanapi_harvesters.auxiliary.path import resolve_rel_path
from ckanapi_harvesters.builder.builder_errors import RequiredDataFrameFieldsError, ResourceFileNotExistMessage, IncompletePatchError
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanDataStoreInfo
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, find_duplicates, datastore_id_col
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice
from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC
from ckanapi_harvesters.builder.mapper_datastore_multi import RequestMapperABC
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreABC
from ckanapi_harvesters.builder.builder_resource_datastore_multi_folder import BuilderDataStoreFolder
from ckanapi_harvesters.builder.builder_resource_multi_abc import BuilderMultiABC, FileChunkDataFrame


class BuilderDataStoreFile(BuilderDataStoreFolder):
    """
    Implementation supporting the reading of a file by chunks
    """
    def __init__(self, *, name: str = None, format: str = None, description: str = None,
                 resource_id: str = None, download_url: str = None, file_name: str = None, options_string:str=None, base_dir:str=None):
        super().__init__(name=name, format=format, description=description, resource_id=resource_id,
                         download_url=download_url, dir_name="", options_string=options_string, base_dir=base_dir)
        self.file_size:int = 0
        self.upsert_method: UpsertChoice = UpsertChoice.Upsert
        self.file_name = file_name

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderDataStoreFile()
        super().copy(dest=dest)
        dest.file_name = self.file_name
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self.file_name: str = _string_from_element(row["file/url"], strip=True)

    @staticmethod
    def resource_mode_str() -> str:
        return "DataStore from File"

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.file_name
        return d

    # def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
    #     file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
    #     if os.path.isfile(file_path):
    #         return None
    #     else:
    #         return ResourceFileNotExistMessage(self.name, ErrorLevel.Error, f"Missing file for resource {self.name}: {file_path}")

    ## upload chunks ----------------
    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
        if os.path.isfile(file_path):
            return None
        else:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Error, f"Missing file for resource {self.name}: {file_path}")

    def get_sample_file_path(self, resources_base_dir:str, ckan:Union[CkanApi,None]=None, file_index:int=0) -> str:
        return resolve_rel_path(resources_base_dir, self.file_name, field=f"File/URL of resource {self.name}")

    def list_local_files(self, resources_base_dir:str, ckan:CkanApi, cancel_if_present:bool=True) -> List[str]:
        file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
        self.file_size = os.path.getsize(file_path)
        self.local_file_list = [file_path]
        self.local_file_size = [self.file_size]
        self.local_file_size_sum = self.file_size
        return self.local_file_list

    def get_local_file_offset(self, file_chunk: FileChunkDataFrame) -> int:
        return file_chunk.file_position

    def get_local_file_total_size(self) -> int:
        return self.file_size

    def get_local_file_count(self) -> int:
        return 1

    ## download ----------------
    def download_request(self, ckan: CkanApi, out_dir: str, *, full_download:bool=True,
                         force:bool=False, threads:int=1, return_data:bool=False) -> Union[pd.DataFrame,None]:
        if (not self.enable_download) and (not force):
            msg = f"Did not download resource {self.name} because download was disabled."
            warn(msg)
            return None
        if out_dir is not None:
            self.downloaded_destination = resolve_rel_path(out_dir, self.file_name, field=f"File/URL of resource {self.name}")
            if self.download_skip_existing and os.path.exists(self.downloaded_destination):
                return None
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and not self.download_error_not_found:
            return None
        if self.local_file_format.append_allowed() and not return_data:
            download_generator = ckan.datastore_dump_generator(resource_id, search_all=full_download)
            first_df = None
            for df_download in download_generator:
                df = self.df_mapper.df_download_alter(df_download, fields=self._get_fields_info())
                if out_dir is not None:
                    if first_df is None:
                        self.local_file_format.write_file(df, self.downloaded_destination, fields=self._get_fields_info())
                        first_df = df  # maybe the first DataFrame could be used in the append function to reproduce treatments
                    else:
                        self.local_file_format.append_file(df, self.downloaded_destination, fields=self._get_fields_info())
        else:
            df_download = ckan.datastore_dump_generator(resource_id, search_all=full_download)
            df = self.df_mapper.df_download_alter(df_download, fields=self._get_fields_info())
            if out_dir is not None:
                os.makedirs(out_dir, exist_ok=True)
                self.local_file_format.write_file(df, self.downloaded_destination, fields=self._get_fields_info())
            if return_data:
                return df
        return None

    def download_request_full(self, ckan: CkanApi, out_dir: str, threads:int=1, external_stop_event=None,
                              start_index:int=0, end_index:int=None, force:bool=False) -> None:
        return self.download_request(ckan=ckan, out_dir=out_dir, full_download=True, force=force)

    def to_builder_datastore_folder(self,
                                    *, dir_name:str=None, primary_key:List[str]=None,
                                    file_query_list:Collection[Tuple[str,dict]]=None) -> BuilderDataStoreFolder:
        resource_folder = BuilderDataStoreFolder()
        resource_folder._load_from_df_row(self._to_row())
        resource_folder.options_string = self.options_string
        resource_folder.data_cleaner_upload = self.data_cleaner_upload.copy()
        resource_folder.field_builders = self.field_builders
        if dir_name is not None:
            resource_folder.dir_name = dir_name
        elif isinstance(self, BuilderDataStoreFolder):
            resource_folder.dir_name = self.dir_name
        else:
            resource_folder.dir_name, _ = os.path.splitext(self.file_name)
        resource_folder.package_name = self.package_name
        if isinstance(self.df_mapper, RequestMapperABC):
            resource_folder.df_mapper = self.df_mapper.copy()
        else:
            resource_folder.df_mapper.df_upload_fun = self.df_mapper.df_upload_fun
            resource_folder.df_mapper.df_download_fun = self.df_mapper.df_download_fun
        if primary_key is not None or file_query_list is not None:
            resource_folder.setup_default_file_mapper(primary_key=primary_key, file_query_list=file_query_list)
        resource_folder.downloaded_file_query_list = file_query_list
        return resource_folder


#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore.
"""
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Union, Set, Any, Generator, Collection
import os
import io
from warnings import warn
from collections import OrderedDict
import copy

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.list_records import ListRecords, GeneralDataFrame
from ckanapi_harvesters.builder.builder_field import BuilderField
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC
from ckanapi_harvesters.harvesters.file_formats.file_format_init import init_file_format_datastore
from ckanapi_harvesters.builder.mapper_datastore import DataSchemeConversion
from ckanapi_harvesters.builder.builder_resource import BuilderResourceABC
from ckanapi_harvesters.auxiliary.ckan_errors import DuplicateNameError
from ckanapi_harvesters.auxiliary.path import resolve_rel_path
from ckanapi_harvesters.builder.builder_errors import RequiredDataFrameFieldsError, ResourceFileNotExistMessage, IncompletePatchError
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanDataStoreInfo
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, find_duplicates, datastore_id_col
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice
from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC
from ckanapi_harvesters.builder.mapper_datastore_multi import RequestMapperABC
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreABC
from ckanapi_harvesters.builder.builder_resource_datastore_multi_folder import BuilderDataStoreFolder
from ckanapi_harvesters.builder.builder_resource_multi_abc import BuilderMultiABC, FileChunkDataFrame


class BuilderDataStoreCkan(BuilderDataStoreFolder):
    """
    Merge of existing CKAN DataStores (on the same server) into a single DataStore
    """
    def __init__(self, *, name: str = None, format: str = None, description: str = None,
                 resource_id: str = None, download_url: str = None, file_name: str = None, options_string:str=None, base_dir:str=None):
        super().__init__(name=name, format=format, description=description, resource_id=resource_id,
                         download_url=download_url, dir_name="", options_string=options_string, base_dir=base_dir)
        self.upsert_method: UpsertChoice = UpsertChoice.Upsert
        self.resource_ids: List[str] = [resource_id.strip() for resource_id in ckan_tags_sep.split(file_name)]
        raise NotImplementedError()  # not tested

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderDataStoreCkan()
        super().copy(dest=dest)
        dest.resource_ids = self.resource_ids
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        file_name_field: str = _string_from_element(row["file/url"], strip=True)
        if file_name_field is not None:
            self.resource_ids = [resource_id.strip() for resource_id in ckan_tags_sep.split(file_name_field)]
        else:
            self.resource_ids = []

    @staticmethod
    def resource_mode_str() -> str:
        return "CKAN DataStore merge"

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return True

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = ckan_tags_sep.join(self.resource_ids)
        return d

    ## upload chunks ----------------
    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        # check resource ids exist on CKAN
        missing_resources = []
        for resource_id in self.resource_ids:
            if ckan.get_resource_info_or_request(resource_id, error_not_found=False) is None:
                missing_resources.append(resource_id)
        if len(missing_resources) == 0:
            return None
        else:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Error, f"Missing source resources for {self.name}: {','.join(missing_resources)}")

    def get_sample_file_path(self, resources_base_dir:str, ckan:Union[CkanApi,None]=None, file_index:int=0) -> str:
        return None

    def list_local_files(self, resources_base_dir:str, ckan:CkanApi, cancel_if_present:bool=True) -> List[str]:
        # obtain row count of each resource
        self.local_file_size = [0] * len(self.resource_ids)
        for index, resource_id in enumerate(self.resource_ids):
            resource_info = ckan.get_resource_info_or_request(resource_id, error_not_found=False, datastore_info=True)
            if resource_info is not None and resource_info.datastore_info is not None and resource_info.datastore_info.row_count is not None:
                self.local_file_size[index] = resource_info.datastore_info.row_count
        self.local_file_size_sum = sum(self.local_file_size,0)
        return self.local_file_list

    def _update_metadata(self, ckan: CkanApi, *, base_dir:str=None):
        # obtain metadata from first resource of list
        super()._update_metadata(ckan, base_dir=base_dir)
        resource_info = ckan.get_resource_info_or_request(self.resource_ids[0], error_not_found=True, datastore_info=True)
        self.resource_attributes_data_source = None
        self.field_builders_data_source = None
        if resource_info is not None:
            self.resource_attributes_data_source = resource_info.copy()
            if resource_info.datastore_info is not None and resource_info.datastore_info.fields_dict is not None:
                self.field_builders_data_source = OrderedDict()
                for field_info in resource_info.datastore_info.fields_dict:
                    self.field_builders_data_source[field_info.field_name] = BuilderField._from_ckan_field(field_info)

    def get_local_df_chunk_generator(self, resources_base_dir:str, ckan:CkanApi, allow_chunks:bool=True,
                                     **kwargs) -> Generator[FileChunkDataFrame, None, None]:
        for file_index, resource_id in enumerate(self.resource_ids):
            self.file_semaphore.acquire()
            file_position = 0
            chunk_index = 0
            generator = ckan.datastore_search_generator(resource_id=resource_id, search_all=True)
            self.file_semaphore.release()
            while True:
                try:
                    self.file_semaphore.acquire()
                    df = next(generator)
                    chunk_index += 1
                    file_position += len(df)
                    self.read_line_counter += len(df)
                    line_counter = self.read_line_counter
                    self.file_semaphore.release()
                except StopIteration:
                    self.file_semaphore.release()
                    return
                yield FileChunkDataFrame(df, resource_id, file_index, chunk_index,
                                         file_position=file_position, read_line_counter=line_counter)





#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore without uploading any data.
"""
import time
from abc import ABC, abstractmethod
from typing import Dict, List, Callable, Any, Tuple, Union, Set, Generator
import os
import io
from warnings import warn

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanProgressCallback
from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.builder.builder_resource import builder_request_default_auth_if_ckan, initial_resource_building_state
from ckanapi_harvesters.builder.builder_resource_datastore_file import BuilderDataStoreFile
from ckanapi_harvesters.builder.builder_resource_multi_abc import FileChunkDataFrame
from ckanapi_harvesters.auxiliary.ckan_errors import NotMappedObjectNameError, DataStoreNotFoundError
from ckanapi_harvesters.auxiliary.list_records import ListRecords, GeneralDataFrame
from ckanapi_harvesters.builder.builder_errors import RequiredDataFrameFieldsError, ResourceFileNotExistMessage
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanDataStoreInfo
from ckanapi_harvesters.auxiliary.ckan_errors import CkanArgumentError, FunctionMissingArgumentError, ExternalUrlLockedError
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, assert_or_raise, find_duplicates, datastore_id_col
from ckanapi_harvesters.ckan_api.ckan_api_2_readonly import df_download_read_csv_kwargs


class BuilderDataStoreUrl(BuilderDataStoreFile):  #, BuilderUrlABC):  # multiple inheritance can give undefined results
    """
    Class representing a DataStore (resource metadata and fields metadata) defined by a url.
    """
    def __init__(self, *, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, url:str=None, options_string:str=None, base_dir:str=None):
        super(BuilderDataStoreFile, self).__init__(name=name, format=format, description=description, resource_id=resource_id,
                                                   download_url=download_url, options_string=options_string, base_dir=base_dir)
        # super(BuilderUrlABC, self).__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url, url=url)
        self.reupload_on_update = False
        self.reupload_if_needed = False
        self.url:str = url
        self.file_name = name

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderDataStoreUrl()
        super().copy(dest=dest)
        dest.reupload_on_update = self.reupload_on_update
        dest.reupload_if_needed = self.reupload_if_needed
        dest.url = self.url
        dest.file_name = self.file_name
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super(BuilderDataStoreFile, self)._load_from_df_row(row=row)
        # super(BuilderUrlABC, self)._load_from_df_row(row=row)
        self.url: str = _string_from_element(row["file/url"], strip=True)
        self.file_name = self.name

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return True

    def get_sample_file_path(self, resources_base_dir: str, ckan:Union[CkanApi,None]=None, file_index:int=0) -> str:
        return self.url

    def init_local_files_list(self, resources_base_dir:str, cancel_if_present:bool=True, **kwargs) -> List[str]:
        file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
        self.file_size = 1
        self.local_file_list = [file_path]
        self.local_file_size = [self.file_size]
        self.local_file_size_sum = 1
        return self.local_file_list

    def load_sample_data(self, resources_base_dir:str, *, ckan:CkanApi=None,
                         proxies:dict=None, headers:dict=None) -> bytes:
        self.sample_source = self.url
        if ckan is None:
            raise FunctionMissingArgumentError("BuilderDataStoreUrl.load_sample_data", "ckan")
        return ckan.download_url_proxy(self.url, proxies=proxies, headers=headers, auth_if_ckan=builder_request_default_auth_if_ckan).content

    def get_local_df_chunk_generator(self, resources_base_dir:str, ckan:CkanApi, **kwargs) -> Generator[FileChunkDataFrame, None, None]:
        payload = self.load_sample_data(resources_base_dir=resources_base_dir)
        buffer = io.StringIO(payload.decode())
        self.read_line_counter = 0
        with self.local_file_format.read_buffer_full(buffer, fields=self._get_fields_info()) as df_file:
            if isinstance(df_file, pd.DataFrame) or isinstance(df_file, ListRecords):
                self.file_semaphore.acquire()
                self.read_line_counter += len(df_file)
                line_counter = self.read_line_counter
                self.file_semaphore.release()
                yield FileChunkDataFrame(df_file, self.url, 0, 0, 0, line_counter)
            else:  # iterator
                with df_file as df_generator:
                    if hasattr(df_file, "handles"):
                        file_handle = df_file.handles.handle
                    else:
                        file_handle = None
                    previous_file_position = 0
                    # for chunk_index, df in enumerate(df_generator):
                    chunk_index = 0
                    file_position = 0
                    while True:
                        self.file_semaphore.acquire()
                        if file_handle is not None:
                            file_position = file_handle.buffer.tell()  # approximative position in file
                        else:
                            file_position = 0  # no position tracking available
                        try:
                            df = next(df_generator)
                            if file_handle is None and hasattr(df, "attrs") and "file_position" in df.attrs:
                                file_position = df.attrs["file_position"]
                            self.read_line_counter += len(df)
                            line_counter = self.read_line_counter
                        except StopIteration:
                            self.file_semaphore.release()
                            break
                        self.file_semaphore.release()
                        yield FileChunkDataFrame(df, self.url, 0, chunk_index, previous_file_position, line_counter)
                        previous_file_position = file_position
                        chunk_index = chunk_index + 1

    def upload_request_full(self, ckan:CkanApi, resources_base_dir:str, *,
                            threads:int=1, external_stop_event=None,
                            start_index:int=0, end_index:int=None, **kwargs) -> None:
        # CKAN manages URL resources: no upload is necessary
        return

    @staticmethod
    def resource_mode_str() -> str:
        return "DataStore from URL"

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.url
        return d

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        if ckan is None:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Warning, "Could not determine if resource url exists because ckan argument was not provided.")
        else:
            return ckan.download_url_proxy_test_head(self.url, **kwargs)

    def patch_request(self, ckan: CkanApi, package_id: str, *,
                      df_upload:pd.DataFrame=None, payload:Union[bytes, io.BufferedIOBase]=None,
                      reupload: bool = None, override_ckan:bool=False, resources_base_dir:str=None) -> CkanResourceInfo:
        """
        Specific implementation of patch_request which does not upload any data and only updates the fields currently present in the database
        :param resources_base_dir:
        :param ckan:
        :param package_id:
        :param reupload:
        :return:
        """
        self._merge_resource_attributes(override_ckan=override_ckan)
        if reupload is None: reupload = self.reupload_on_update
        if payload is not None or df_upload is not None:
            raise CkanArgumentError("payload", "datastore defined from URL patch")
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=False)
        try:
            df_download = self.download_sample_df(ckan, download_alter=False, search_all=False, limit=1)
            if df_download is None:
                assert_or_raise(resource_id is None, RuntimeError("Unexpected: resource_id should be None"))
                raise NotMappedObjectNameError(self.name)
            current_df_fields = set(df_download.columns)
        except NotMappedObjectNameError as e:
            df_download = None
            current_df_fields = set()
        except DataStoreNotFoundError as e:
            df_download = None
            current_df_fields = set()
        empty_datastore = df_download is None or len(df_download) == 0
        data_cleaner_fields = None
        data_cleaner_index = set()
        current_df_fields -= {datastore_id_col}  # _id does not require documentation
        aliases = self._get_alias_list(ckan)
        self._check_necessary_fields(current_df_fields, raise_error=False, empty_datastore=empty_datastore)
        self._check_undocumented_fields(current_df_fields)
        primary_key, indexes = self._get_primary_key_indexes(data_cleaner_index, current_df_fields=current_df_fields,
                                                             error_missing=False, empty_datastore=empty_datastore)
        fields_update = self._get_fields_update(ckan, current_df_fields=current_df_fields, data_cleaner_fields=data_cleaner_fields,
                                                reupload=reupload, override_ckan=override_ckan)
        fields = list(fields_update.values()) if len(fields_update) > 0 else None
        resource_info = ckan.resource_create(package_id, name=self.name, format=self.resource_attributes.format, description=self.resource_attributes.description,
                                             state=initial_resource_building_state if initial_resource_building_state is not None else self.resource_attributes.state,
                                             url=self.url,
                                             datastore_create=False, auto_submit=False, create_default_view=self.create_default_view,
                                             cancel_if_exists=True, update_if_exists=True, aliases=aliases, reupload=False, data_cleaner=self.data_cleaner_upload)
        resource_id = resource_info.id
        self.known_id = resource_id
        self._compare_fields_to_datastore_info(resource_info, current_df_fields, ckan)
        if reupload:
            # re-initialize datastore to reupload from url
            # normally, data was automatically submitted to DataStore on resource_create (not needed)
            ckan.datastore_create(resource_id, fields=fields, primary_key=primary_key, indexes=indexes, aliases=aliases)
            ckan.datastore_submit(resource_id)
        return resource_info






#!python3
# -*- coding: utf-8 -*-
"""
Progress callback function definition
"""
from typing import Any, Union, Callable, Set
from enum import IntEnum
import copy

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_auxiliary import FileChunkDataFrame


class CkanCallbackLevel(IntEnum):
    Packages = 0
    Resources = 1
    ResourceChunks = 2
    MultiFileResource = 3
    Request = 4


def default_progress_callback(position:int, total:int, info:Any=None, *, context:str=None,
                              file_index:int=None, file_count:int=None,
                              lines_chunk:int=None, total_lines_read:int=None,
                              canceled_upload: bool=False, end_message: bool=False, level:CkanCallbackLevel=None,
                              **kwargs) -> None:
    if level == CkanCallbackLevel.Packages:
        print(f"Package {position}/{total}")
    elif level == CkanCallbackLevel.Resources:
        if end_message:
            print(f"Done all {total} resources")
        else:
            print(f"Resource {position}/{total}")
    elif level == CkanCallbackLevel.MultiFileResource:
        print(f"Multi-file resource / file {position}/{total}")
    elif level == CkanCallbackLevel.ResourceChunks:
        if context is None:
            context = ""
        if position == total and end_message:
            # info is None
            print(f"{context} Finished {file_index}/{file_count} (100%) - {total_lines_read} lines read")
        elif canceled_upload:
            print(f"{context} Canceled {file_index}/{file_count} ({position/total*100.0:.2f}%) - {total_lines_read} lines read")
        elif info is None:
            print(f"{context} Request {file_index}/{file_count} ({position/total*100.0:.2f}%) - {total_lines_read} lines read")
        else:
            if isinstance(info, str):
                info_str = info
            elif isinstance(info, pd.DataFrame):
                if "source" in info.attrs.keys():
                    info_str = str(info.attrs["source"])
                else:
                    info_str = "<DataFrame>"
            elif isinstance(info, list):
                info_str = "<records>"
            elif isinstance(info, FileChunkDataFrame):
                if isinstance(info.df, pd.DataFrame):
                    if "source" in info.df.attrs.keys():
                        info_str = str(info.df.attrs["source"])
                    else:
                        info_str = "<DataFrame>"
                elif isinstance(info.df, list):
                    info_str = "<records>"
                else:
                    info_str = str(info.df)
            else:
                info_str = str(info)
            print(f"{context} Request {file_index}/{file_count} ({position/total*100.0:.2f}%) - {total_lines_read} lines read: " + info_str)
    elif level == CkanCallbackLevel.Request:
        print(f"Multi-line request {position}/{total} ({position/total*100.0:.2f}%)")


class CkanProgressCallback:
    def __init__(self, callback_fun:Union[Callable,"CkanProgressCallback"]=None):
        self.progress_callback_fun: Union[Callable[[int, int, Any], None], None] = None
        self.progress_callback_kwargs: dict = {}
        self.verbosity:dict[CkanCallbackLevel,bool] = {level: True for level in CkanCallbackLevel}
        # self.verbosity[CkanCallbackLevel.Request] = False
        if isinstance(callback_fun, CkanProgressCallback):
            callback_fun.copy(dest=self)
            return
        else:
            if callback_fun is None:
                callback_fun = default_progress_callback
            self.progress_callback_fun = callback_fun

    def copy(self, *, dest=None):
        dest = CkanProgressCallback()
        dest.progress_callback_fun = self.progress_callback_fun
        dest.progress_callback_kwargs = copy.deepcopy(self.progress_callback_kwargs)
        dest.verbosity = self.verbosity.copy()
        return dest

    def __copy__(self):
        return self.copy()

    def call(self, position:int, total:int, *, info:Any=None, context:str=None,
             file_index:int=0, file_count:int=None, lines_chunk:int=None, total_lines_read:int=None,
             canceled_request: bool=False, end_message: bool=False, level:CkanCallbackLevel=0,
             **kwargs) -> None:
        """
        Progress callback function. Use to implement a progress indication for the user.

        :param position: the position within the resource (usually, in bytes or line count)
        :param total: the total size of the resource
        :param info: an object from which more information can be extracted, typically, the DataFrame itself, with an indication of the data origin.
        :param context: the context of the call (ckan instance, upload/download, single/multi-threaded)
        :param file_index: the index of the file in the list
        :param file_count: the number of files in the list
        :param lines_chunk: the number of lines in the chunk currently being processed
        :param total_lines_read: the total number of lines read, including the current chunk
        :param canceled_request: this callback is also called when a line is ignored
        :param end_message: boolean indicating of the work in progress
        :param level: the level of the progress callback
        """
        if self.progress_callback_fun is not None:
            if not self.verbosity[level]:
                return
            if end_message:
                position = total
                file_index = file_count
            self.progress_callback_fun(position, total, info=info, context=context,
                                       file_index=file_index, file_count=file_count,
                                       lines_chunk=lines_chunk, total_lines_read=total_lines_read,
                                       canceled_upload=canceled_request, end_message=end_message,
                                       level=level, **self.progress_callback_kwargs, **kwargs)



#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing
"""
from typing import List, Any, Iterable, Union, Dict, Set
from abc import ABC, abstractmethod
import re
import fnmatch
from collections import OrderedDict

from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element
from ckanapi_harvesters.policies.data_format_policy_errors import DataPolicyError, ErrorLevel, _policy_msg
from ckanapi_harvesters.policies.data_format_policy_defs import StringMatchMode, ListChoiceMode, newline_char
from ckanapi_harvesters.policies.data_format_policy_defs import StringValueSpecification


class DataPolicyABC(ABC):
    def __init__(self, error_level:ErrorLevel=ErrorLevel.Information):
        self.error_level: ErrorLevel = error_level

    @abstractmethod
    def to_dict(self) -> dict:
        return {"error_level": self.error_level.name}

    @staticmethod
    @abstractmethod
    def from_dict(d:dict):
        raise NotImplementedError()

    @abstractmethod
    def _load_from_dict(self, d:dict):
        self.error_level = ErrorLevel.from_str(d["error_level"])

    @abstractmethod
    def enforce(self, values: Any, *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        raise NotImplementedError()


class DataPolicyElementABC(DataPolicyABC, ABC):
    def __init__(self, mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(error_level=error_level)
        self.mandatory:bool = mandatory
        self.match_mode: StringMatchMode = StringMatchMode.Match

    @abstractmethod
    def to_dict(self) -> dict:
        d = OrderedDict([("mandatory", self.mandatory), ("match_mode", self.match_mode.name)])
        d.update(super().to_dict())
        return d

    @staticmethod
    @abstractmethod
    def from_dict(d:dict):
        raise NotImplementedError()

    @abstractmethod
    def _load_from_dict(self, d:dict) -> None:
        super()._load_from_dict(d)
        self.mandatory = d["mandatory"]
        self.match_mode = StringMatchMode.from_str(d["match_mode"])

    def _enforce_unit_string(self, values: Union[str, List[str]], spec: Union[str, Iterable[str]], *, context:str, verbose:bool,
                             buffer:List[DataPolicyError], add_buffer:bool=True) -> bool:
        if values is None or len(values) == 0:
            return not self.match_mode == StringMatchMode.NotEmpty
        if isinstance(values, str):
            values = [values]
        success = True
        for value in values:
            success_value = True
            if isinstance(spec, str):
                spec = [spec]
            if value is None:
                if self.mandatory:
                    success_value = False
            elif self.match_mode == StringMatchMode.Match:
                success_value = value.lower() in {unit_spec.lower() for unit_spec in spec}
            elif self.match_mode == StringMatchMode.MatchCaseSensitive:
                success_value = value in spec
            elif self.match_mode == StringMatchMode.Regex:
                # TODO: test
                success_value = any([re.match(unit_spec,value,flags=re.IGNORECASE) is not None for unit_spec in spec])
            elif self.match_mode == StringMatchMode.RegexCaseSensitive:
                # TODO: test
                success_value = any([re.match(unit_spec,value) is not None for unit_spec in spec])
            elif self.match_mode == StringMatchMode.Wildcard:
                success_value = any([fnmatch.fnmatch(value, unit_spec) is not None for unit_spec in spec])
            elif self.match_mode == StringMatchMode.WildcardCaseSensitive:
                success_value = any([fnmatch.fnmatchcase(value, unit_spec) is not None for unit_spec in spec])
            if add_buffer and not success_value:
                msg = DataPolicyError(context, self.error_level, f"Value does not match spec '{spec}' ({self.match_mode}): {value}")
                _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
            success &= success_value
        return success



